/*
 * Copyright (C) 2022 Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */

#include "config.h"

#include "grd-credentials.h"

G_DEFINE_ABSTRACT_TYPE (GrdCredentials, grd_credentials, G_TYPE_OBJECT)

GVariant *
grd_credentials_lookup (GrdCredentials      *credentials,
                        GrdCredentialsType   type,
                        GError             **error)
{
  return GRD_CREDENTIALS_GET_CLASS (credentials)->lookup (credentials,
                                                          type, error);
}

gboolean
grd_credentials_clear (GrdCredentials      *credentials,
                       GrdCredentialsType   type,
                       GError             **error)
{
  return GRD_CREDENTIALS_GET_CLASS (credentials)->clear (credentials,
                                                         type, error);
}

gboolean
grd_credentials_store (GrdCredentials      *credentials,
                       GrdCredentialsType   type,
                       GVariant            *variant,
                       GError             **error)
{
  return GRD_CREDENTIALS_GET_CLASS (credentials)->store (credentials,
                                                         type, variant, error);
}

static void
grd_credentials_init (GrdCredentials *credentials)
{
}

static void
grd_credentials_class_init (GrdCredentialsClass *klass)
{
}
